﻿#region License Information

/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  [Item("RegressionSolution Impacts Calculator", "Calculation of the impacts of input variables for any regression solution")]
  public sealed class RegressionSolutionVariableImpactsCalculator : ParameterizedNamedItem {
    public enum ReplacementMethodEnum {
      Median,
      Average,
      Shuffle,
      Noise
    }
    public enum FactorReplacementMethodEnum {
      Best,
      Mode,
      Shuffle
    }
    public enum DataPartitionEnum {
      Training,
      Test,
      All
    }

    private const string ReplacementParameterName = "Replacement Method";
    private const string DataPartitionParameterName = "DataPartition";

    public IFixedValueParameter<EnumValue<ReplacementMethodEnum>> ReplacementParameter
    {
      get { return (IFixedValueParameter<EnumValue<ReplacementMethodEnum>>)Parameters[ReplacementParameterName]; }
    }
    public IFixedValueParameter<EnumValue<DataPartitionEnum>> DataPartitionParameter
    {
      get { return (IFixedValueParameter<EnumValue<DataPartitionEnum>>)Parameters[DataPartitionParameterName]; }
    }

    public ReplacementMethodEnum ReplacementMethod
    {
      get { return ReplacementParameter.Value.Value; }
      set { ReplacementParameter.Value.Value = value; }
    }
    public DataPartitionEnum DataPartition
    {
      get { return DataPartitionParameter.Value.Value; }
      set { DataPartitionParameter.Value.Value = value; }
    }


    [StorableConstructor]
    private RegressionSolutionVariableImpactsCalculator(bool deserializing) : base(deserializing) { }
    private RegressionSolutionVariableImpactsCalculator(RegressionSolutionVariableImpactsCalculator original, Cloner cloner)
      : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new RegressionSolutionVariableImpactsCalculator(this, cloner);
    }

    public RegressionSolutionVariableImpactsCalculator()
      : base() {
      Parameters.Add(new FixedValueParameter<EnumValue<ReplacementMethodEnum>>(ReplacementParameterName, "The replacement method for variables during impact calculation.", new EnumValue<ReplacementMethodEnum>(ReplacementMethodEnum.Median)));
      Parameters.Add(new FixedValueParameter<EnumValue<DataPartitionEnum>>(DataPartitionParameterName, "The data partition on which the impacts are calculated.", new EnumValue<DataPartitionEnum>(DataPartitionEnum.Training)));
    }

    //mkommend: annoying name clash with static method, open to better naming suggestions
    public IEnumerable<Tuple<string, double>> Calculate(IRegressionSolution solution) {
      return CalculateImpacts(solution, DataPartition, ReplacementMethod);
    }

    private static void PrepareData(DataPartitionEnum partition,
      IRegressionSolution solution,
      out IEnumerable<int> rows,
      out IEnumerable<double> targetValues,
      out double originalR2) {
      OnlineCalculatorError error;

      switch (partition) {
        case DataPartitionEnum.All:
          rows = solution.ProblemData.AllIndices;
          targetValues = solution.ProblemData.TargetVariableValues.ToList();
          originalR2 = OnlinePearsonsRCalculator.Calculate(solution.ProblemData.TargetVariableValues, solution.EstimatedValues, out error);
          if (error != OnlineCalculatorError.None) throw new InvalidOperationException("Error during R² calculation.");
          originalR2 = originalR2 * originalR2;
          break;
        case DataPartitionEnum.Training:
          rows = solution.ProblemData.TrainingIndices;
          targetValues = solution.ProblemData.TargetVariableTrainingValues.ToList();
          originalR2 = solution.TrainingRSquared;
          break;
        case DataPartitionEnum.Test:
          rows = solution.ProblemData.TestIndices;
          targetValues = solution.ProblemData.TargetVariableTestValues.ToList();
          originalR2 = solution.TestRSquared;
          break;
        default: throw new ArgumentException(string.Format("DataPartition {0} cannot be handled.", partition));
      }
    }

    private static double CalculateImpactForDouble(string variableName,
      IRegressionSolution solution,
      ModifiableDataset modifiableDataset,
      IEnumerable<int> rows,
      IEnumerable<double> targetValues,
      double originalR2,
      ReplacementMethodEnum replacementMethod) {
      OnlineCalculatorError error;
      var newEstimates = EvaluateModelWithReplacedVariable(solution.Model, variableName, modifiableDataset, rows, replacementMethod);
      var newR2 = OnlinePearsonsRCalculator.Calculate(targetValues, newEstimates, out error);
      if (error != OnlineCalculatorError.None) { throw new InvalidOperationException("Error during R² calculation with replaced inputs."); }
      return originalR2 - (newR2 * newR2);
    }

    private static double CalculateImpactForString(string variableName,
      IRegressionSolution solution,
      ModifiableDataset modifiableDataset,
      IEnumerable<int> rows,
      IEnumerable<double> targetValues,
      double originalR2,
      FactorReplacementMethodEnum factorReplacementMethod) {

      OnlineCalculatorError error;
      if (factorReplacementMethod == FactorReplacementMethodEnum.Best) {
        // try replacing with all possible values and find the best replacement value
        var smallestImpact = double.PositiveInfinity;
        foreach (var repl in solution.ProblemData.Dataset.GetStringValues(variableName, rows).Distinct()) {
          var newEstimates = EvaluateModelWithReplacedVariable(solution.Model, variableName, modifiableDataset, rows, Enumerable.Repeat(repl, solution.ProblemData.Dataset.Rows));
          var newR2 = OnlinePearsonsRCalculator.Calculate(targetValues, newEstimates, out error);
          if (error != OnlineCalculatorError.None) throw new InvalidOperationException("Error during R² calculation with replaced inputs.");

          var curImpact = originalR2 - (newR2 * newR2);
          if (curImpact < smallestImpact) smallestImpact = curImpact;
        }
        return smallestImpact;
      } else {
        // for replacement methods shuffle and mode
        // calculate impacts for factor variables
        var newEstimates = EvaluateModelWithReplacedVariable(solution.Model, variableName, modifiableDataset, rows, factorReplacementMethod);
        var newR2 = OnlinePearsonsRCalculator.Calculate(targetValues, newEstimates, out error);
        if (error != OnlineCalculatorError.None) throw new InvalidOperationException("Error during R² calculation with replaced inputs.");

        return originalR2 - (newR2 * newR2);
      }
    }
    public static double CalculateImpact(string variableName,
      IRegressionSolution solution,
      IEnumerable<int> rows,
      IEnumerable<double> targetValues,
      double originalR2,
      DataPartitionEnum data = DataPartitionEnum.Training,
      ReplacementMethodEnum replacementMethod = ReplacementMethodEnum.Median,
      FactorReplacementMethodEnum factorReplacementMethod = FactorReplacementMethodEnum.Best) {

      double impact = 0;
      var modifiableDataset = ((Dataset)solution.ProblemData.Dataset).ToModifiable();

      // calculate impacts for double variables
      if (solution.ProblemData.Dataset.VariableHasType<double>(variableName)) {
        impact = CalculateImpactForDouble(variableName, solution, modifiableDataset, rows, targetValues, originalR2, replacementMethod);
      } else if (solution.ProblemData.Dataset.VariableHasType<string>(variableName)) {
        impact = CalculateImpactForString(variableName, solution, modifiableDataset, rows, targetValues, originalR2, factorReplacementMethod);
      } else {
        throw new NotSupportedException("Variable not supported");
      }
      return impact;
    }

    public static IEnumerable<Tuple<string, double>> CalculateImpacts(
      IRegressionSolution solution,
      DataPartitionEnum data = DataPartitionEnum.Training,
      ReplacementMethodEnum replacementMethod = ReplacementMethodEnum.Median,
      FactorReplacementMethodEnum factorReplacementMethod = FactorReplacementMethodEnum.Best,
      Func<double, string, bool> progressCallback = null) {

      IEnumerable<int> rows;
      IEnumerable<double> targetValues;
      double originalR2 = -1;

      PrepareData(data, solution, out rows, out targetValues, out originalR2);

      var impacts = new Dictionary<string, double>();
      var inputvariables = new HashSet<string>(solution.ProblemData.AllowedInputVariables.Union(solution.Model.VariablesUsedForPrediction));
      var allowedInputVariables = solution.ProblemData.Dataset.VariableNames.Where(v => inputvariables.Contains(v)).ToList();

      int curIdx = 0;
      int count = allowedInputVariables.Where(solution.ProblemData.Dataset.VariableHasType<double>).Count();
      // calculate impacts for double variables
      foreach (var inputVariable in allowedInputVariables) {
        //Report the current progress in percent. If the callback returns true, it means the execution shall be stopped
        if (progressCallback != null) {
          curIdx++;
          if (progressCallback((double)curIdx / count, string.Format("Calculating impact for variable {0} ({1} of {2})", inputVariable, curIdx, count))) { return null; }
        }
        impacts[inputVariable] = CalculateImpact(inputVariable, solution, rows, targetValues, originalR2, data, replacementMethod, factorReplacementMethod);
      }

      return impacts.OrderByDescending(i => i.Value).Select(i => Tuple.Create(i.Key, i.Value));
    }

    private static IEnumerable<double> EvaluateModelWithReplacedVariable(IRegressionModel model, string variable, ModifiableDataset dataset, IEnumerable<int> rows, ReplacementMethodEnum replacement = ReplacementMethodEnum.Median) {
      var originalValues = dataset.GetReadOnlyDoubleValues(variable).ToList();
      double replacementValue;
      List<double> replacementValues;
      IRandom rand;

      switch (replacement) {
        case ReplacementMethodEnum.Median:
          replacementValue = rows.Select(r => originalValues[r]).Median();
          replacementValues = Enumerable.Repeat(replacementValue, dataset.Rows).ToList();
          break;
        case ReplacementMethodEnum.Average:
          replacementValue = rows.Select(r => originalValues[r]).Average();
          replacementValues = Enumerable.Repeat(replacementValue, dataset.Rows).ToList();
          break;
        case ReplacementMethodEnum.Shuffle:
          // new var has same empirical distribution but the relation to y is broken
          rand = new FastRandom(31415);
          // prepare a complete column for the dataset
          replacementValues = Enumerable.Repeat(double.NaN, dataset.Rows).ToList();
          // shuffle only the selected rows
          var shuffledValues = rows.Select(r => originalValues[r]).Shuffle(rand).ToList();
          int i = 0;
          // update column values 
          foreach (var r in rows) {
            replacementValues[r] = shuffledValues[i++];
          }
          break;
        case ReplacementMethodEnum.Noise:
          var avg = rows.Select(r => originalValues[r]).Average();
          var stdDev = rows.Select(r => originalValues[r]).StandardDeviation();
          rand = new FastRandom(31415);
          // prepare a complete column for the dataset
          replacementValues = Enumerable.Repeat(double.NaN, dataset.Rows).ToList();
          // update column values 
          foreach (var r in rows) {
            replacementValues[r] = NormalDistributedRandom.NextDouble(rand, avg, stdDev);
          }
          break;

        default:
          throw new ArgumentException(string.Format("ReplacementMethod {0} cannot be handled.", replacement));
      }

      return EvaluateModelWithReplacedVariable(model, variable, dataset, rows, replacementValues);
    }

    private static IEnumerable<double> EvaluateModelWithReplacedVariable(
      IRegressionModel model, string variable, ModifiableDataset dataset,
      IEnumerable<int> rows,
      FactorReplacementMethodEnum replacement = FactorReplacementMethodEnum.Shuffle) {
      var originalValues = dataset.GetReadOnlyStringValues(variable).ToList();
      List<string> replacementValues;
      IRandom rand;

      switch (replacement) {
        case FactorReplacementMethodEnum.Mode:
          var mostCommonValue = rows.Select(r => originalValues[r])
            .GroupBy(v => v)
            .OrderByDescending(g => g.Count())
            .First().Key;
          replacementValues = Enumerable.Repeat(mostCommonValue, dataset.Rows).ToList();
          break;
        case FactorReplacementMethodEnum.Shuffle:
          // new var has same empirical distribution but the relation to y is broken
          rand = new FastRandom(31415);
          // prepare a complete column for the dataset
          replacementValues = Enumerable.Repeat(string.Empty, dataset.Rows).ToList();
          // shuffle only the selected rows
          var shuffledValues = rows.Select(r => originalValues[r]).Shuffle(rand).ToList();
          int i = 0;
          // update column values 
          foreach (var r in rows) {
            replacementValues[r] = shuffledValues[i++];
          }
          break;
        default:
          throw new ArgumentException(string.Format("FactorReplacementMethod {0} cannot be handled.", replacement));
      }

      return EvaluateModelWithReplacedVariable(model, variable, dataset, rows, replacementValues);
    }

    private static IEnumerable<double> EvaluateModelWithReplacedVariable(IRegressionModel model, string variable,
      ModifiableDataset dataset, IEnumerable<int> rows, IEnumerable<double> replacementValues) {
      var originalValues = dataset.GetReadOnlyDoubleValues(variable).ToList();
      dataset.ReplaceVariable(variable, replacementValues.ToList());
      //mkommend: ToList is used on purpose to avoid lazy evaluation that could result in wrong estimates due to variable replacements
      var estimates = model.GetEstimatedValues(dataset, rows).ToList();
      dataset.ReplaceVariable(variable, originalValues);

      return estimates;
    }
    private static IEnumerable<double> EvaluateModelWithReplacedVariable(IRegressionModel model, string variable,
      ModifiableDataset dataset, IEnumerable<int> rows, IEnumerable<string> replacementValues) {
      var originalValues = dataset.GetReadOnlyStringValues(variable).ToList();
      dataset.ReplaceVariable(variable, replacementValues.ToList());
      //mkommend: ToList is used on purpose to avoid lazy evaluation that could result in wrong estimates due to variable replacements
      var estimates = model.GetEstimatedValues(dataset, rows).ToList();
      dataset.ReplaceVariable(variable, originalValues);

      return estimates;
    }
  }
}
